//  Lars Grimsland - lars.grimsland@gmail.com
//  NMPC with wind disturbance

#include "short_NLP.hpp"
#include "parameters.h"

#include <cassert>
#include <iostream>
#include <fstream>

//#include <math.h>
#include <cmath>

// translating matlab symbols used in included file
#define x10 x[10+i*8]
#define x11 x[11+i*8]
#define x12 x[12+i*8]
#define x13 x[13+i*8]
#define x14 x[14+i*8]
#define x15 x[15+i*8]

#define x0  x[0+i*8] // 0,6,12,18....
#define x1  x[1+i*8]
#define x2  x[2+i*8]
#define x3  x[3+i*8]
#define x4  x[4+i*8]
#define x5  x[5+i*8]
#define x6  x[6+i*8]
#define x7  x[7+i*8]
#define x8  x[8+i*8]
#define x9  x[9+i*8]

#define lambda0 lambda[0+i*6]
#define lambda1 lambda[1+i*6]
#define lambda2 lambda[2+i*6]
#define lambda3 lambda[3+i*6]
#define lambda4 lambda[4+i*6]
#define lambda5 lambda[5+i*6]
#define lambda6 lambda[6+i*6]

#define symfix 1


using namespace Ipopt;

// constructor
short_NLP::short_NLP()
{}

//destructor
short_NLP::~short_NLP()
{}

// returns the size of the problem
bool short_NLP::get_nlp_info(Index& n, Index& m, Index& nnz_jac_g,
                             Index& nnz_h_lag, IndexStyleEnum& index_style)
{
	std::cout <<  "Get NLP Info" << std::endl;
	n = get_noVar();	// parameters defined in parameters.cpp
	m = get_noConstr();
	
	nnz_jac_g = get_conJacNonZ(); // number of nonzeros in jacobian of constraints
	nnz_h_lag = get_lagrHessNonZ(); // nonzeros in lower left corner og lagrang hessian
	
	// use the C style indexing (0-based)
	index_style = TNLP::C_STYLE;
	
	return true;
}

// returns the variable bounds
bool short_NLP::get_bounds_info(Index n, Number* x_l, Number* x_u,
                                Index m, Number* g_l, Number* g_u)
{
	// here, the n and m we gave IPOPT in get_nlp_info are passed back to us.
	// If desired, we could assert to make sure they are what we think they are.
	
	int limit=get_segments();
	assert(n == get_noVar());
	assert(m == get_noConstr());

	// the variables have lower bounds of 1
	double lower_var_bounds[8] = {-2e19, -2e19, 11.2, -2e19, -2e19, -2e19, -3.00, -0.523598};
	//                            North  East   Speed  Dir     Unc   Unc    Acc     Bank

	// lower (and upper) constraints on the first node is used to fix a startpoint for the flight path 
    x_l[0] = get_startPosUavNorth();		// lower_var_bounds[0];
	x_l[1] = get_startPosUavEast();		// lower_var_bounds[1];
	x_l[2] = get_startUavSpeed();	// lower_var_bounds[2];
	x_l[3] = get_startUavHeading();		//lower_var_bounds[3];
	x_l[4] = get_startValUncertainty0();		//lower_var_bounds[4];
	x_l[5] = get_startValUncertainty1();
	x_l[6] = lower_var_bounds[6];
	x_l[7] = lower_var_bounds[7];

	
	for(int i=0; i<limit; i++){
		x_l[ 8+i*8]  = lower_var_bounds[0]; //x_l[ 6, 12, 18...] 	
		x_l[ 9+i*8]  = lower_var_bounds[1];
		x_l[10+i*8] = lower_var_bounds[2];
		x_l[11+i*8] = lower_var_bounds[3];
		x_l[12+i*8] = lower_var_bounds[4];
		x_l[13+i*8] = lower_var_bounds[5];
		x_l[14+i*8] = lower_var_bounds[6];
		x_l[15+i*8] = lower_var_bounds[7];
	}

  // the variables have upper bounds
	double upper_var_bounds[8] = {2e19, 2e19, 26.8, 2e19, 2e19, 2e19, 3.00, 0.523598};
    //                           North  East  Speed Dir   Unc   Unc    Acc    Bank

    x_u[0] = get_startPosUavNorth();		
	x_u[1] = get_startPosUavEast();	
	x_u[2] = get_startUavSpeed();		
	x_u[3] = get_startUavHeading();			
	x_u[4] = get_startValUncertainty0();	
	x_u[5] = get_startValUncertainty1();	
	x_u[6] = upper_var_bounds[6];
	x_u[7] = upper_var_bounds[7];
	
	for(int i=0; i< limit; i++){
		x_u[ 8+i*8] = upper_var_bounds[0]; //x_u[ 6, 12, 18...]
		x_u[ 9+i*8] = upper_var_bounds[1];
		x_u[10+i*8] = upper_var_bounds[2];
		x_u[11+i*8] = upper_var_bounds[3];
		x_u[12+i*8] = upper_var_bounds[4];
		x_u[13+i*8] = upper_var_bounds[5];
		x_u[14+i*8] = upper_var_bounds[6];
		x_u[15+i*8] = upper_var_bounds[7];
	}
  

   // the second constraint g2 is an equality constraint, so we set the
  // upper and lower bound to the same value
  
  for(int i=0; i<(m) ;i++){
	g_l[i] = 0; // all constraints are equality constraints
	g_u[i] = 0;
  }

  return true;
}

// returns the initial point for the problem
bool short_NLP::get_starting_point(Index n, bool init_x, Number* x,
                                   bool init_z, Number* z_L, Number* z_U,
                                   Index m, bool init_lambda,
                                   Number* lambda)
{
  // Here, we assume we only have starting values for x, if you code
  // your own NLP, you can provide starting values for the dual variables
  // if you wish
	
	int limit=get_segments();
	double  T = get_T();
	
	double	alpha0 = get_alpha0();
	double	b0 = get_b0();
	double	k0 = get_k0();
	double ceta0 = get_ceta0();
	double	T0n= get_T0N();	// returns north coordinate of grid center 0
	double	T0e= get_T0E();	// returns east coordinate of grid center 0
	
	double	alpha1 = get_alpha1();
	double	b1 = get_b1();
	double	k1 = get_k1();
	double ceta1 = get_ceta1();
	double	T1n= get_T1N();	// returns north coordinate of grid center 1
	double	T1e= get_T1E();	// returns east coordinate of grid center 1
	
	double xRes[31*8];
	
	assert(init_x == true);
	assert(init_z == false);
	assert(init_lambda == false);
	
	
	//if (get_loopTest() == 0) {
		// values in the first node
		x[0] = get_startPosUavNorth();
		x[1] = get_startPosUavEast();
		x[2] = get_startUavSpeed();
		x[3] = get_startUavHeading();
		x[4] = get_startValUncertainty0();
		x[5] = get_startValUncertainty1();
		x[6] = 0; //inputs are zero to ease start point calculation
		x[7] = 0; 
		// initialize to the given starting point, 
		for(int i=0; i<get_segments(); i++){
			#include "C:\textfiles\startval.txt"
		}
//	}
/* 	else{
		/* Importing results of last optimization */
		
//		#include "C:\Users\grimslan\Dropbox\ShortFlight - 2ExpBell NMPC wind\Release\xRes.txt"
//		std::cout << x[1] << "  <--- xRes x[1] er  .... <--   " << std::endl; 
//		std::cout << x[2] << "  <--- xRes x[2] er  .... <--   " << std::endl; 
//		/* "Moving" the solution from previous MPC to the next startpoint for optimizing */
//		for(int i=0; i<get_OptVarsNode()* ( get_nodes() - get_recalcNodeNo() ); i++ ){
//			x[i]=xRes[i+ get_OptVarsNode()*get_recalcNodeNo() ];
//		}
//		/* Simulating a flight through the last node(s) - get_recalcNodeNo() nodes! 
//		   startval.txt contains: x[8+i*8]=... ,x[9+i*8]=... ,...,x[15+i*8]=...          */
//		for(int i=( get_OptVarsNode()*( get_nodes() - get_recalcNodeNo()-1)  ); i<get_nodes()-1; i++){
//			#include "C:\textfiles\startval.txt"
//		}
//	}
 
	return true;
}

// returns the value of the objective function
bool short_NLP::eval_f(Index n, const Number* x, bool new_x, Number& obj_value)
{
	//std::cout <<  "Get Eval F" << std::endl;
	assert(n == get_noVar());
	double Wa = get_Wa(); // returns wheigting (punishments) on input 1, acceleration
	double Wr = get_Wr(); // returns wheigting (punishments) on input 2, rudder/bank angle
	double Wu0 = get_Wu0(); // returns wheigting (punishments) on UNCERTAINTY
	double Wu1 = get_Wu1();

	#include "C:\textfiles\objective.txt"
	return true;
}

// return the gradient of the objective function grad_{x} f(x)
bool short_NLP::eval_grad_f(Index n, const Number* x, bool new_x, Number* grad_f)
{
	assert(n == get_noVar()); //nodes*system_variables);
	double Wa = get_Wa(); // returns wheigting (punishments) on input 1, acceleration
	double Wr = get_Wr(); // returns wheigting (punishments) on input 2, rudder/bank angle
	double Wu0 = get_Wu0(); // returns wheigting (punishments) on UNCERTAINTY
	double Wu1 = get_Wu1();
	
	#include "C:\textfiles\obj_jac.txt"
		
	return true;
}

// return the value of the constraints: g(x)
bool short_NLP::eval_g(Index n, const Number* x, bool new_x, Index m, Number* g)
{
	//std::cout << " constraints value begin  \n" ;
	assert(n == get_noVar());
	assert(m == get_noConstr());
	double T = get_T();
	double grav = get_grav();

	double alpha0 =	get_alpha0();
	double b0	= get_b0();
	double k0	= get_k0();
	double ceta0= get_ceta0();
	double T0n	= get_T0N();
	double T0e	= get_T0E();
	
	double alpha1=	get_alpha0();
	double b1	= get_b0();
	double k1	= get_k0();
	double ceta1= get_ceta1();
	double T1n	= get_T1N();
	double T1e	= get_T1E();
		
	int i=0;
	int limit = get_segments();
	for(i=0; i<limit; i++){
			#include "C:\textfiles\g_ccode.txt"
	}
	return true;
}

// return the structure or values of the constraints jacobian
bool short_NLP::eval_jac_g(Index n, const Number* x, bool new_x,
                           Index m, Index nele_jac, Index* iRow, Index *jCol,
                           Number* values)
{
	int i=0;
	int limit = get_segments();
	
	if (values == NULL) {   // return the structure of the jacobian
		
		for(i=0; i<limit; i++){
				#include "C:\textfiles\JacPos.txt"
		}
  }
  else {
    // return the values of the jacobian of the constraints
	double T	 = get_T();
	double grav  = get_grav();

	double alpha0 = get_alpha0();
	double b0	 = get_b0();
	double k0	 = get_k0();
	double ceta0 = get_ceta0();
	double T0n	 = get_T0N();
	double T0e	 = get_T0E();

	double alpha1 = get_alpha1();
	double b1	 = get_b1();
	double k1	 = get_k1();
	double ceta1 = get_ceta1();
	double T1n	 = get_T1N();
	double T1e	 = get_T1E();
	
    
	for(i=0; i<limit; i++){
			#include "C:\textfiles\JacVal.txt"
	}
  }
	//std::cout << "done with jacobian \n";
	return true;
}

//return the structure or values of the hessian
bool short_NLP::eval_h(Index n, const Number* x, bool new_x,
                       Number obj_factor, Index m, const Number* lambda,
                       bool new_lambda, Index nele_hess, Index* iRow,
                       Index* jCol, Number* values)
{
	//std::cout << "Hessian begun \n";
	double T	= get_T();			// segment length
	double grav	= get_grav();

	double alpha0 = get_alpha0();
	double b0	  = get_b0();
	double k0	  = get_k0();// various parameter in
	double ceta0  = get_ceta0(); // uncertainty model
	double T0n	  = get_T0N();		
	double T0e	  = get_T0E();

	double alpha1 = get_alpha1();
	double b1	  = get_b1();
	double k1	  = get_k1();// various parameter in
	double ceta1  = get_ceta1(); // uncertainty model
	double T1n	  = get_T1N();		
	double T1e	  = get_T1E();
		
	double Wa	= get_Wa();		// wheigt, acceleration
	double Wr	= get_Wr();		// wheigt, bank angle
	double Wu0	= get_Wu0();		// wheigt, UNCERTAINTY
	double Wu1	= get_Wu1();
	/*Prep for Hessian block generalization*/
	double valuebase[140];
	int v_index;


	if (values == NULL) {
		// return the structure. This is a symmetric matrix, fill the lower left
		// triangle only.
		#include "C:\textfiles\hesspos.txt"
  }
  else {
    // return the values. This is a symmetric matrix, fill the lower left triangle only
		
		/*Including a single Hessian Block , generalizing values*/
	
		for(int i=0; i<get_segments(); i++ ){		// i couts segments
			#include "C:\textfiles\hessval2.txt"
			for(int nn=0; nn<get_lagrHessNonzBlock(); nn++){
				if(i==0){
					/*No appending in first block*/
					values[nn]=valuebase[nn];
				}	
				else if (i>0 && nn<get_HessNoOverlaps() ){
					v_index=(i-1)*(get_lagrHessNonzBlock()-get_HessNoOverlaps())+get_HessOverlapIndex(nn);
					values[v_index]=values[v_index]+valuebase[nn];
				}
				else{
				v_index = nn+(get_lagrHessNonzBlock()-get_HessNoOverlaps())*i;
				values[v_index]=valuebase[nn];
				}
			} // for loop counting Hess Nonzeros nn ends
		} // for loop counting segments, i ends
  }
  return true;
}

void short_NLP::finalize_solution(SolverReturn status,
                                  Index n, const Number* x, const Number* z_L, const Number* z_U,
                                  Index m, const Number* g, const Number* lambda,
                                  Number obj_value,
								  const IpoptData* ip_data,
								  IpoptCalculatedQuantities* ip_cq)
{
  // here is where we would store the solution to variables, or write to a file, etc
  // so we could use the solution.

	int node = get_recalcNodeNo();
	int shift = get_OptVarsNode();


	/* "Correcting position" - simulating a measurement of wind disturbance */
	double xCorr[ 31*8 ];
	for(Index i=0; i<n; i++){
		xCorr[i]=x[i];
	}
	for (int i=0; i<get_nodes(); i++){
		/* NOTE only north position coordinate affected*/
		xCorr[0+i*get_OptVarsNode()]=xCorr[i*get_OptVarsNode()]-get_T()*i*get_windSpeedNorth();
	}
	/* Writing corrected solution to file, so it can be used as start point for next optimization*/
	std::ofstream xRes;
	xRes.open("xRes.txt");
	for (Index i=0; i<n; i++) {
		xRes <<"xRes["<<i<<"]="	<< xCorr[i] <<";"<< std::endl;
	}
	xRes.close();

	/* Writing solution of optimization to file */ 	
	if (get_loopTest()==0){
		/* First iteration of MPC - write whole "first" optimization to a file */
		std::ofstream firstoptim;
		std::ofstream firstoptimWind;
		firstoptimWind.open( "first_optim_with_disturbance.txt" );
		firstoptim.open( "first_optim.txt" );
		for (Index i=0; i<n; i++) {
			firstoptimWind	<< xCorr[i] << std::endl;
			firstoptim		<< x[i]		<< std::endl;
		}
		firstoptimWind.close();
		firstoptim.close();
	}

	/* If first iteration of MPC - write various parameters to file */
	if (get_loopTest()==0){
		
		std::ofstream parameter_file;
		parameter_file.open ("parameters.txt");
		
		/* Problem data */
		parameter_file << get_StateVarsNode() << std::endl;
		parameter_file << get_InputVarsNode() << std::endl;
		parameter_file << get_T()			  << std::endl;
		parameter_file << get_recalcNodeNo()  << std::endl;
		// done writing parameteres, closing file
		parameter_file.close();
	}
	
	std::ofstream solutionfile;
	if (get_loopTest() >0){
		//unless first NMPC iteration, data is appended in file
		solutionfile.open( "opt_solution.txt", std::ios::out | std::ios::app );
	}
	else{
		//on first NMPC iteration, file is flushed before writing
		solutionfile.open ("opt_solution.txt");
		
	}
	
	if(get_mpcLoopNr() <= get_noMpcIter()+2 ){ // pluss to added as quickfix to NT write whole path on last MMPC recalc
		// print only "flewn" nodes
		for (Index i=0; i<shift*node; i++) {
			solutionfile << xCorr[i] << std::endl;
		}
	}
/*	else{
		// last iteration of NMPC loop, print whole optimal path
		for (Index i=0; i<n; i++) {
			solutionfile << xCorr[i] << std::endl;
		}
	}
*/
	solutionfile.close();

	/* Setting startpoint for next optimization */
	/* Simulating a drift o southwards due to wind */
	set_startPosUavNorth( xCorr[0+node*shift] ); 
	set_startPosUavEast(  x[1+node*shift] );
	set_startUavSpeed(    x[2+node*shift] );
	set_startUavHeading(  x[3+node*shift] );
	set_startValUncertainty0( x[4+node*shift] );
	set_startValUncertainty1( x[5+node*shift] );

//   Writing info to console
  std::cout << std::endl << std::endl << "Objective value" << std::endl;
  std::cout << "f(x*) = " << obj_value << std::endl;

  if(get_loopTest() ==0){
	set_loopTest(2);
  }

}
